@@ -178,11 +178,23 @@ class TestNodeWaitForProvisionState(base.TestCase):
 
         mock_fetch.side_effect = _get_side_effect
 
-        self.assertRaisesRegex(exceptions.SDKException,
+        self.assertRaisesRegex(exceptions.ResourceFailure,
                                'failure state "deploy failed"',
                                self.node.wait_for_provision_state,
                                self.session, 'manageable')
 
+    def test_failure_error(self, mock_fetch):
+        def _get_side_effect(_self, session):
+            self.node.provision_state = 'error'
+            self.assertIs(session, self.session)
+
+        mock_fetch.side_effect = _get_side_effect
+
+        self.assertRaisesRegex(exceptions.ResourceFailure,
+                               'failure state "error"',
+                               self.node.wait_for_provision_state,
+                               self.session, 'manageable')
+
     def test_enroll_as_failure(self, mock_fetch):
         def _get_side_effect(_self, session):
             self.node.provision_state = 'enroll'
@@ -191,7 +203,7 @@ class TestNodeWaitForProvisionState(base.TestCase):
 
         mock_fetch.side_effect = _get_side_effect
 
-        self.assertRaisesRegex(exceptions.SDKException,
+        self.assertRaisesRegex(exceptions.ResourceFailure,
                                'failed to verify management credentials',
                                self.node.wait_for_provision_state,
                                self.session, 'manageable')
@@ -225,7 +237,54 @@ class TestNodeSetProvisionState(base.TestCase):
                                  default_microversion=None)
 
     def test_no_arguments(self):
-        self.node.set_provision_state(self.session, 'manage')
+        result = self.node.set_provision_state(self.session, 'active')
+        self.assertIs(result, self.node)
+        self.session.put.assert_called_once_with(
+            'nodes/%s/states/provision' % self.node.id,
+            json={'target': 'active'},
+            headers=mock.ANY, microversion=None,
+            retriable_status_codes=_common.RETRIABLE_STATUS_CODES)
+
+    def test_manage(self):
+        result = self.node.set_provision_state(self.session, 'manage')
+        self.assertIs(result, self.node)
+        self.session.put.assert_called_once_with(
+            'nodes/%s/states/provision' % self.node.id,
+            json={'target': 'manage'},
+            headers=mock.ANY, microversion='1.4',
+            retriable_status_codes=_common.RETRIABLE_STATUS_CODES)
+
+    def test_deploy_with_configdrive(self):
+        result = self.node.set_provision_state(self.session, 'active',
+                                               config_drive='abcd')
+        self.assertIs(result, self.node)
+        self.session.put.assert_called_once_with(
+            'nodes/%s/states/provision' % self.node.id,
+            json={'target': 'active', 'configdrive': 'abcd'},
+            headers=mock.ANY, microversion=None,
+            retriable_status_codes=_common.RETRIABLE_STATUS_CODES)
+
+    def test_rebuild_with_configdrive(self):
+        result = self.node.set_provision_state(self.session, 'rebuild',
+                                               config_drive='abcd')
+        self.assertIs(result, self.node)
+        self.session.put.assert_called_once_with(
+            'nodes/%s/states/provision' % self.node.id,
+            json={'target': 'rebuild', 'configdrive': 'abcd'},
+            headers=mock.ANY, microversion='1.35',
+            retriable_status_codes=_common.RETRIABLE_STATUS_CODES)
+
+    def test_configdrive_as_dict(self):
+        for target in ('rebuild', 'active'):
+            self.session.put.reset_mock()
+            result = self.node.set_provision_state(
+                self.session, target, config_drive={'user_data': 'abcd'})
+            self.assertIs(result, self.node)
+            self.session.put.assert_called_once_with(
+                'nodes/%s/states/provision' % self.node.id,
+                json={'target': target, 'configdrive': {'user_data': 'abcd'}},
+                headers=mock.ANY, microversion='1.56',
+                retriable_status_codes=_common.RETRIABLE_STATUS_CODES)
 
 
 @mock.patch.object(node.Node, '_translate_response', mock.Mock())
@@ -250,7 +309,8 @@ class TestNodeCreate(base.TestCase):
         self.assertIs(result, self.node)
         self.session.post.assert_called_once_with(
             mock.ANY, json={'driver': FAKE['driver']},
-            headers=mock.ANY, microversion=self.session.default_microversion)
+            headers=mock.ANY, microversion=self.session.default_microversion,
+            params={})
         self.assertFalse(mock_prov.called)
 
     def test_available_new_version(self, mock_prov):
@@ -266,7 +326,8 @@ class TestNodeCreate(base.TestCase):
         self.assertIs(result, self.node)
         self.session.post.assert_called_once_with(
             mock.ANY, json={'driver': FAKE['driver']},
-            headers=mock.ANY, microversion=self.session.default_microversion)
+            headers=mock.ANY, microversion=self.session.default_microversion,
+            params={})
         mock_prov.assert_has_calls([
             mock.call(self.node, self.session, 'manage', wait=True),
             mock.call(self.node, self.session, 'provide', wait=True)
@@ -288,7 +349,8 @@ class TestNodeCreate(base.TestCase):
         self.assertIs(result, self.node)
         self.session.post.assert_called_once_with(
             mock.ANY, json={'driver': FAKE['driver']},
-            headers=mock.ANY, microversion=self.session.default_microversion)
+            headers=mock.ANY, microversion=self.session.default_microversion,
+            params={})
         self.assertFalse(mock_prov.called)
 
     def test_no_manageable_in_old_version(self, mock_prov):
@@ -307,7 +369,8 @@ class TestNodeCreate(base.TestCase):
         self.assertIs(result, self.node)
         self.session.post.assert_called_once_with(
             mock.ANY, json={'driver': FAKE['driver']},
-            headers=mock.ANY, microversion=self.session.default_microversion)
+            headers=mock.ANY, microversion=self.session.default_microversion,
+            params={})
         mock_prov.assert_called_once_with(self.node, self.session, 'manage',
                                           wait=True)
 
@@ -320,7 +383,8 @@ class TestNodeCreate(base.TestCase):
         self.assertIs(result, self.node)
         self.session.post.assert_called_once_with(
             mock.ANY, json={'driver': FAKE['driver']},
-            headers=mock.ANY, microversion=self.session.default_microversion)
+            headers=mock.ANY, microversion=self.session.default_microversion,
+            params={})
         mock_prov.assert_called_once_with(self.node, self.session, 'manage',
                                           wait=True)
 
@@ -333,6 +397,7 @@ class TestNodeVif(base.TestCase):
         super(TestNodeVif, self).setUp()
         self.session = mock.Mock(spec=adapter.Adapter)
         self.session.default_microversion = '1.28'
+        self.session.log = mock.Mock()
         self.node = node.Node(id='c29db401-b6a7-4530-af8e-20a720dee946',
                               driver=FAKE['driver'])
         self.vif_id = '714bdf6d-2386-4b5e-bd0d-bc036f04b1ef'
@@ -458,6 +523,7 @@ class TestNodeWaitForReservation(base.TestCase):
         super(TestNodeWaitForReservation, self).setUp()
         self.session = mock.Mock(spec=adapter.Adapter)
         self.session.default_microversion = '1.6'
+        self.session.log = mock.Mock()
         self.node = node.Node(**FAKE)
 
     def test_no_reservation(self, mock_fetch):
@@ -487,3 +553,214 @@ class TestNodeWaitForReservation(base.TestCase):
                           self.node.wait_for_reservation,
                           self.session, timeout=0.001)
         mock_fetch.assert_called_with(self.node, self.session)
+
+
+@mock.patch.object(exceptions, 'raise_from_response', mock.Mock())
+class TestNodeSetPowerState(base.TestCase):
+
+    def setUp(self):
+        super(TestNodeSetPowerState, self).setUp()
+        self.node = node.Node(**FAKE)
+        self.session = mock.Mock(spec=adapter.Adapter,
+                                 default_microversion=None)
+
+    def test_power_on(self):
+        self.node.set_power_state(self.session, 'power on')
+        self.session.put.assert_called_once_with(
+            'nodes/%s/states/power' % FAKE['uuid'],
+            json={'target': 'power on'},
+            headers=mock.ANY,
+            microversion=None,
+            retriable_status_codes=_common.RETRIABLE_STATUS_CODES)
+
+    def test_soft_power_on(self):
+        self.node.set_power_state(self.session, 'soft power off')
+        self.session.put.assert_called_once_with(
+            'nodes/%s/states/power' % FAKE['uuid'],
+            json={'target': 'soft power off'},
+            headers=mock.ANY,
+            microversion='1.27',
+            retriable_status_codes=_common.RETRIABLE_STATUS_CODES)
+
+
+@mock.patch.object(exceptions, 'raise_from_response', mock.Mock())
+@mock.patch.object(node.Node, '_translate_response', mock.Mock())
+@mock.patch.object(node.Node, '_get_session', lambda self, x: x)
+class TestNodeMaintenance(base.TestCase):
+
+    def setUp(self):
+        super(TestNodeMaintenance, self).setUp()
+        self.node = node.Node.existing(**FAKE)
+        self.session = mock.Mock(spec=adapter.Adapter,
+                                 default_microversion='1.1',
+                                 retriable_status_codes=None)
+
+    def test_set(self):
+        self.node.set_maintenance(self.session)
+        self.session.put.assert_called_once_with(
+            'nodes/%s/maintenance' % self.node.id,
+            json={'reason': None},
+            headers=mock.ANY,
+            microversion=mock.ANY)
+
+    def test_set_with_reason(self):
+        self.node.set_maintenance(self.session, 'No work on Monday')
+        self.session.put.assert_called_once_with(
+            'nodes/%s/maintenance' % self.node.id,
+            json={'reason': 'No work on Monday'},
+            headers=mock.ANY,
+            microversion=mock.ANY)
+
+    def test_unset(self):
+        self.node.unset_maintenance(self.session)
+        self.session.delete.assert_called_once_with(
+            'nodes/%s/maintenance' % self.node.id,
+            json=None,
+            headers=mock.ANY,
+            microversion=mock.ANY)
+
+    def test_set_via_update(self):
+        self.node.is_maintenance = True
+        self.node.commit(self.session)
+        self.session.put.assert_called_once_with(
+            'nodes/%s/maintenance' % self.node.id,
+            json={'reason': None},
+            headers=mock.ANY,
+            microversion=mock.ANY)
+
+        self.assertFalse(self.session.patch.called)
+
+    def test_set_with_reason_via_update(self):
+        self.node.is_maintenance = True
+        self.node.maintenance_reason = 'No work on Monday'
+        self.node.commit(self.session)
+        self.session.put.assert_called_once_with(
+            'nodes/%s/maintenance' % self.node.id,
+            json={'reason': 'No work on Monday'},
+            headers=mock.ANY,
+            microversion=mock.ANY)
+        self.assertFalse(self.session.patch.called)
+
+    def test_set_with_other_fields(self):
+        self.node.is_maintenance = True
+        self.node.name = 'lazy-3000'
+        self.node.commit(self.session)
+        self.session.put.assert_called_once_with(
+            'nodes/%s/maintenance' % self.node.id,
+            json={'reason': None},
+            headers=mock.ANY,
+            microversion=mock.ANY)
+
+        self.session.patch.assert_called_once_with(
+            'nodes/%s' % self.node.id,
+            json=[{'path': '/name', 'op': 'replace', 'value': 'lazy-3000'}],
+            headers=mock.ANY,
+            microversion=mock.ANY)
+
+    def test_set_with_reason_and_other_fields(self):
+        self.node.is_maintenance = True
+        self.node.maintenance_reason = 'No work on Monday'
+        self.node.name = 'lazy-3000'
+        self.node.commit(self.session)
+        self.session.put.assert_called_once_with(
+            'nodes/%s/maintenance' % self.node.id,
+            json={'reason': 'No work on Monday'},
+            headers=mock.ANY,
+            microversion=mock.ANY)
+
+        self.session.patch.assert_called_once_with(
+            'nodes/%s' % self.node.id,
+            json=[{'path': '/name', 'op': 'replace', 'value': 'lazy-3000'}],
+            headers=mock.ANY,
+            microversion=mock.ANY)
+
+    def test_no_reason_without_maintenance(self):
+        self.node.maintenance_reason = 'Can I?'
+        self.assertRaises(ValueError, self.node.commit, self.session)
+        self.assertFalse(self.session.put.called)
+        self.assertFalse(self.session.patch.called)
+
+    def test_set_unset_maintenance(self):
+        self.node.is_maintenance = True
+        self.node.maintenance_reason = 'No work on Monday'
+        self.node.commit(self.session)
+
+        self.session.put.assert_called_once_with(
+            'nodes/%s/maintenance' % self.node.id,
+            json={'reason': 'No work on Monday'},
+            headers=mock.ANY,
+            microversion=mock.ANY)
+
+        self.node.is_maintenance = False
+        self.node.commit(self.session)
+        self.assertIsNone(self.node.maintenance_reason)
+
+        self.session.delete.assert_called_once_with(
+            'nodes/%s/maintenance' % self.node.id,
+            json=None,
+            headers=mock.ANY,
+            microversion=mock.ANY)
+
+
+@mock.patch.object(node.Node, 'fetch', lambda self, session: self)
+@mock.patch.object(exceptions, 'raise_from_response', mock.Mock())
+class TestNodeSetBootDevice(base.TestCase):
+
+    def setUp(self):
+        super(TestNodeSetBootDevice, self).setUp()
+        self.node = node.Node(**FAKE)
+        self.session = mock.Mock(spec=adapter.Adapter,
+                                 default_microversion='1.1')
+
+    def test_node_set_boot_device(self):
+        self.node.set_boot_device(self.session, 'pxe', persistent=False)
+        self.session.put.assert_called_once_with(
+            'nodes/%s/management/boot_device' % self.node.id,
+            json={'boot_device': 'pxe', 'persistent': False},
+            headers=mock.ANY, microversion=mock.ANY,
+            retriable_status_codes=_common.RETRIABLE_STATUS_CODES)
+
+
+@mock.patch.object(node.Node, 'fetch', lambda self, session: self)
+@mock.patch.object(exceptions, 'raise_from_response', mock.Mock())
+class TestNodeTraits(base.TestCase):
+
+    def setUp(self):
+        super(TestNodeTraits, self).setUp()
+        self.node = node.Node(**FAKE)
+        self.session = mock.Mock(spec=adapter.Adapter,
+                                 default_microversion='1.37')
+        self.session.log = mock.Mock()
+
+    def test_node_add_trait(self):
+        self.node.add_trait(self.session, 'CUSTOM_FAKE')
+        self.session.put.assert_called_once_with(
+            'nodes/%s/traits/%s' % (self.node.id, 'CUSTOM_FAKE'),
+            json=None,
+            headers=mock.ANY, microversion='1.37',
+            retriable_status_codes=_common.RETRIABLE_STATUS_CODES)
+
+    def test_remove_trait(self):
+        self.node.remove_trait(self.session, 'CUSTOM_FAKE')
+        self.session.delete.assert_called_once_with(
+            'nodes/%s/traits/%s' % (self.node.id, 'CUSTOM_FAKE'),
+            headers=mock.ANY, microversion='1.37',
+            retriable_status_codes=_common.RETRIABLE_STATUS_CODES)
+
+    def test_remove_trait_missing(self):
+        self.session.delete.return_value.status_code = 400
+        self.assertFalse(self.node.remove_trait(self.session,
+                                                'CUSTOM_MISSING'))
+        self.session.delete.assert_called_once_with(
+            'nodes/%s/traits/%s' % (self.node.id, 'CUSTOM_MISSING'),
+            headers=mock.ANY, microversion='1.37',
+            retriable_status_codes=_common.RETRIABLE_STATUS_CODES)
+
+    def test_set_traits(self):
+        traits = ['CUSTOM_FAKE', 'CUSTOM_REAL', 'CUSTOM_MISSING']
+        self.node.set_traits(self.session, traits)
+        self.session.put.assert_called_once_with(
+            'nodes/%s/traits' % self.node.id,
+            json={'traits': ['CUSTOM_FAKE', 'CUSTOM_REAL', 'CUSTOM_MISSING']},
+            headers=mock.ANY, microversion='1.37',
+            retriable_status_codes=_common.RETRIABLE_STATUS_CODES)
